# -*- coding: utf-8 -*-
"""
Created on Tue Nov 19 22:50:26 2019

@author: yoelr
"""
from flexsolve.jit_speed import njit_alternative
from flexsolve.bounded_solvers import IQ_interpolation
from flexsolve import utils

__all__ = ('secant', 'aitken_secant')

@njit_alternative(cache=True)
def secant(f, x0, x1=None, xtol=0., ytol=5e-8, args=(), maxiter=50,
           checkroot=False, checkiter=True):
    """Secant solver."""
    if checkroot: utils.check_tols(xtol, ytol)
    if x1 is None: x1 = x0 + 1e-5
    y0 = f(x0, *args)
    abs_ = abs
    if abs_(y0) < ytol: return x0
    dx = x1-x0 
    xi0 = x0
    for iter in range(maxiter): 
        y1 = f(x1, *args)
        xtol_satisfied = abs_(dx) < xtol
        ytol_satisfied = abs_(y1) < ytol
        if ytol_satisfied or xtol_satisfied: 
            if checkroot and ytol_satisfied and xtol_satisfied: return x1
            else: return x1
        if y1*y0 < 0.: 
            return IQ_interpolation(f, xi0, x1, y0, y1, None,
                                    xtol, ytol, args, maxiter-iter,
                                    checkroot, checkiter, False)
        if y1 == y0:
            if checkroot or checkiter: utils.raise_tol_error()
            else: return x1
        xi0 = x1
        x1 = x0 - y1*dx/(y1-y0)
        dx = x1-x0 
        x0 = x1
        y0 = y1
    if checkiter: utils.raise_iter_error()
    return x1

@njit_alternative(cache=True)
def aitken_secant(f, x0, x1=None, xtol=0., ytol=5e-8, args=(), maxiter=50,
                  checkroot=False, checkiter=True):
    """Secant solver with Aitken acceleration."""
    if checkroot: utils.check_tols(xtol, ytol)
    if x1 is None: x1 = x0 + 1e-5
    abs_ = abs
    y0 = f(x0, *args)
    if abs_(y0) < ytol: return x0
    dx = x1-x0
    aitken_iter = utils.scalar_aitken_iter
    oscillating = False
    for iter in range(maxiter):
        y1 = f(x1, *args)
        xtol_satisfied = abs_(dx) < xtol
        ytol_satisfied = abs_(y1) < ytol
        if ytol_satisfied or xtol_satisfied: 
            if checkroot and ytol_satisfied and xtol_satisfied: return x1
            else: return x1
        inbounds = y1*y0 < 0.
        if y1 == y0:
            if inbounds:
                return IQ_interpolation(f, x0, x1, y0, y1, None,
                                        xtol, ytol, args, maxiter-iter,
                                        checkroot, False)
            if checkroot or checkiter: utils.raise_tol_error()
            else: return x1
        if inbounds:
            if oscillating:
                return IQ_interpolation(f, x0, x1, y0, y1, None,
                                        xtol, ytol, args, maxiter-iter,
                                        checkroot, False)
            else:
                oscillating = True
            
        x0 = x1 - y1*dx/(y1-y0) # x0 = g
        dx = x0-x1
        y0 = y1
        y1 = f(x0, *args)
        xtol_satisfied = abs_(dx) < xtol
        ytol_satisfied = abs_(y1) < ytol
        if ytol_satisfied or xtol_satisfied: 
            if checkroot and ytol_satisfied and xtol_satisfied: return x0
            else: return x0
        inbounds = y1*y0 < 0.
        if y1 == y0:
            if inbounds:
                return IQ_interpolation(f, x1, x0, y0, y1, None,
                                        xtol, ytol, args, maxiter-iter,
                                        checkroot, False)
            if checkroot or checkiter: utils.raise_tol_error()
            else: return x1
        if inbounds:
            if oscillating:
                return IQ_interpolation(f, x1, x0, y0, y1, None,
                                        xtol, ytol, args, maxiter-iter,
                                        checkroot, False)
            else:
                oscillating = True
        x2 = x0 - y1*dx/(y1-y0) # x2 = gg
        dx = x1 - x0 # x - g
        x1 = aitken_iter(x1, x2, dx, x2 - x0)
        dx = x1 - x0
        y0 = y1
    if checkiter: utils.raise_iter_error()
    return x1


# -*- coding: utf-8 -*-
"""
Created on Thu Jun 25 22:50:56 2020

@author: yrc2
"""
from collections.abc import Mapping
from .problem import Problem
import pandas as pd
import numpy as np

__all__ = ('ProblemList',)

class ProblemList(list):
    
    def add_problem(self, f=None, cases=()):
        if not f: return lambda f: self.add_problem(f, cases)
        problem = Problem(f, cases)
        self.append(problem)
        return problem
    
    def profiles_list(self, solver, kwargs):
        if isinstance(kwargs, Mapping):
            return [i.profile_solver(solver, 1e-10, kwargs) for i in self]
        else:
            return [i.profile_solver(solver, 1e-10, j) for i,j in zip(self, kwargs)]
    
    def profiles_dict(self, solver, kwargs):
        if isinstance(kwargs, Mapping):
            return {i.name: i.profile_solver(solver, 1e-10, kwargs) for i in self}
        else:
            return {i.name: i.profile_solver(solver, 1e-10, j) for i,j in zip(self, kwargs)}
    
    def profiles_dicts(self, solvers, solver_kwargs=None, solver_names=None):
        solver_kwargs = solver_kwargs or {}
        solver_names = solver_names or [i.__name__ for i in solvers]
        return {
            name: self.profiles(solver, kwargs)
            for name, solver, kwargs in zip(solver_names, solvers, solver_kwargs)
        }
        
    def results_df(self, solvers, solver_kwargs=None, solver_names=None):
        solver_problem_profiles = [self.profiles_list(i, j) for i,j in zip(solvers, solver_kwargs)]
        problem_names = [i.name for i in self]
        problem_fields = ['Iterations', 'Passed', 'Failed']
        multi_index = pd.MultiIndex.from_product([problem_names, problem_fields], names=['Problem', 'Summary'])
        columns = solver_names
        get_column = lambda x: sum([(i.size(), len(i.passed_cases), len(i.failed_cases)) for i in x], ())
        data = [get_column(i) for i in solver_problem_profiles]
        data = np.array(data).transpose()
        return pd.DataFrame(data, columns=columns, index=multi_index)
    
    def summary_df(self, solvers, solver_kwargs=None, solver_names=None):
        problem_profiles = [self.profiles_list(i, j) for i,j in zip(solvers, solver_kwargs)]
        passed_cases = [sum([len(i.passed_cases) for i in j]) for j in problem_profiles]
        failed_cases = [sum([len(i.failed_cases) for i in j]) for j in problem_profiles]
        failed_problems = [sum([not i.failed_cases for i in j]) for j in problem_profiles]
        data = np.array([passed_cases, failed_cases, failed_problems])
        return pd.DataFrame(data, columns=solver_names, index=['Passed cases', 'Failed cases', 'Failed problems'])
    
    def __repr__(self):
        return f"{type(self).__name__}({list.__repr__(self)})"


# -*- coding: utf-8 -*-
"""
Created on Thu Jun 25 22:49:17 2020

@author: yrc2
"""
from .iterative_solvers import wegstein, aitken
from .profiler import Profiler

__all__ = ('Problem',)

fixedpoint_solvers = [wegstein, aitken]

class Problem:
    __slots__ = ('f', 'cases')
    
    def __init__(self, f, cases):
        self.f = f
        self.cases = cases
    
    @property
    def name(self):
        return self.f.__name__.capitalize().replace('_', ' ')
        
    def test(self, x, tol, args): 
        assert abs(self.f(x, *args)) <= tol, "result not within tolerance"
        
    def f_fixed_point(self, x, *args):
        return self.f(x, *args) + x
        
    def profile_solver(self, solver, ytol, kwargs={}):
        f = self.f_fixed_point if solver in fixedpoint_solvers else self.f
        p = Profiler(f)
        for case in self.cases:
            try: 
                x = solver(p, 0., **kwargs)
                self.test(x, ytol, kwargs.get('args', ()))
            except: 
                p.archive_case(case, failed=True)
            else:
                p.archive_case(case)
        return p
    
    def __repr__(self):
        return f"{type(self).__name__}(f={self.f}, cases={self.cases})"
    
    def _ipython_display_(self):
        print(f"{type(self).__name__}: {self.name}\n"
              f" cases: {self.cases}")


# -*- coding: utf-8 -*-
"""
Created on Thu Jun 25 19:58:16 2020

@author: yrc2
"""

from math import log, exp, erf, pi, sin, cos
import numpy as np
import flexsolve as flx

test_problems = flx.ProblemList()
add_problem = test_problems.add_problem

### Known problems ###

@add_problem(cases=[6.25 + 5.0, 6.25 - 1.0, 6.25 + 0.1])
def newton_baffler(x):
    x = x - 6.25
    if x < -0.25:
        return 0.75*x- 0.3125
    elif x < 0.25:
        return 2.0*x
    else:
        return 0.75*x + 0.3125

@add_problem(cases=[2, 0.5, 4])
def flat_stanley(x):
    if x == 1:
        return 0
    else:
        factor = (-1.0 if x < 1.0 else 1.0)
        return factor*exp(log(1000) + log(abs(x - 1.0)) - 1.0/(x - 1.0)**2)

@add_problem(cases=[0.01, -0.25])
def newton_pathological(x):
    if x == 0.0:
        return 0.0
    else:
        factor = (-1.0 if x < 0.0 else 1.0)
        return factor*abs(x)**(1.0/3.0)*exp(-x**2)

@add_problem(cases=[1., -0.14, 0.041])
def repeller(x):
     return 20*x/(100*x**2 + 1)
    
@add_problem(cases=[.25, 5., 1.1])
def pinhead(x):
    return (16 - x**4)/(16*x**4 + 0.00001)
    
@add_problem(cases=[100., 1.])
def lazy_boy(x):
    return 0.00000000001*(x - 100)

@add_problem(cases=[0., 5+180., 5.])
def kepler(x):
    return pi*(x - 5.0)/180.0 - 0.8*sin(pi*x/180)

@add_problem(cases=[3., -0.5, 0, 2.12742])
def camel(x):
    return 1.0/((x - 0.3)**2 + 0.01) + 1.0/((x - 0.9)**2 + 0.04) + 2.0*x - 5.2

@add_problem(cases=[2., 3.])
def Wallis_example(x):
    return x**3 - 2*x - 5

### Unnamed ###

@add_problem(cases=[0.1])
def zero_test_1(x):
    return sin(x) - x/2
        
@add_problem(cases=[1.])
def zero_test_2(x):
    return 2*x - exp(-x)
        
@add_problem(cases=[1.])
def zero_test_3(x):
    return x*exp(-x)

@add_problem(cases=[1.])
def zero_test_4(x):
    return exp(x) - 1.0/(10.0*x)**2

@add_problem(cases=[1.])
def zero_test_5(x):
    return (x+3)*(x-1)**2

@add_problem(cases=[1.])
def zero_test_6(x):
    return exp(x) - 2 - 1/(10*x)**2 + 2/(100*x)**3
       
@add_problem(cases=[1.])
def zero_test_7(x):
    return x**3

@add_problem(cases=[1.])
def zero_test_8(x):
    return cos(x) - x

@add_problem(cases=[.99, 1.013])
def zero_test_9(x):
    return 10.0**14*(1.0*x**7 - 7.0*x**6 + 21.0*x**5 - 35.0*x**4 + 35.0*x**3 - 21.0*x**2 + 7.0*x - 1.0)
        
@add_problem(cases=[0., 1., 0.5])
def zero_test_10(x):
    return cos(100.0*x) - 4.0*erf(30*x - 10)
        
### From Scipy ###

@add_problem(cases=[3,])
def scipy_test_1(x):
    return x**2 - 2*x - 1

@add_problem(cases=[3,])
def scipy_test_2(x):
    return exp(x) - cos(x)

@add_problem(cases=[-1e8, 1e7])
def scipy_GH5555(x):
    return x - 0.1

@add_problem(cases=[0., 1.])
def scipy_GH5557(x):
    return -0.1 if x < 0.5 else x - 0.6 # Fail at 0 is expected - 0 slope
        
@add_problem(cases=[10*(200.0 - 6.828499381469512e-06) / (2.0 + 6.828499381469512e-06)])
def zero_der_nz_dp(x):
    return (x - 100.0)**2 
    
@add_problem(cases=[0., 0.5])
def scipy_GH8904(x):
    return x**3 - x**2
        
@add_problem(cases=[0., 0.5])
def scipy_GH8881(x):
    return x**(1.00/9.0) - 9**(1.0/9)
        
### From gsl ###
        
@add_problem(cases=[3, 4, -4, -3, -1/3., 1])
def gsl_test_1(x):
    return sin(x)

@add_problem(cases=[0, 3, -3, 0])
def gsl_test_2(x):
    return cos(x)
        
@add_problem(cases=[0.1, 2])
def gsl_test_3(x):
    return x**20 - 1

@add_problem(cases=[-1.0/3, 1])
def gsl_test_4(x):
    return np.sign(x)*abs(x)**0.5

@add_problem(cases=[0, 1])
def gsl_test_5(x):
    return x**2 - 1e-8
        
@add_problem(cases=[0.9995, 1.0002])
def gsl_test_6(x):
    return (x-1.0)**7
        
### From Roots.jl ###

@add_problem(cases=[0, 1])
def roots_test_1(x):
    return abs(x - 0.0)

@add_problem(cases=[0, -1, 1, 21])
def roots_test_2(x):
    return 1024*x**11 - 2816*x**9 + 2816*x**7 - 1232*x**5 + 220*x**3 - 11*x

@add_problem(cases=[0, -1, 1, 7])
def roots_test_3(x):
    return 512*x**9 - 1024*x**7 + 672*x**5 - 160*x**3 +10*x


if __name__ == '__main__':
    solvers = [flx.secant, flx.aitken_secant, flx.wegstein, flx.aitken]
    solver_names = [i.__name__ for i in solvers]
    kwargs = [{'ytol': 1e-10}, {'ytol': 1e-10}, {'xtol': 1e-10}, {'xtol': 1e-10}]
    df_results = test_problems.results_df(solvers,
                                  solver_kwargs=kwargs,
                                  solver_names=solver_names)
    df_summary = test_problems.summary_df(solvers,
                                  solver_kwargs=kwargs,
                                  solver_names=solver_names)
    